package ru.ifmo.lift;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.util.Random;

import com.evelopers.unimod.runtime.*;
import com.evelopers.unimod.runtime.context.*;
import com.evelopers.unimod.core.stateworks.*;

/**
 * 
 * Lift class. Contains all functionality.
 *
 */
public class Lift implements ControlledObject 
{
	public static final int FLOORS_COUNT = 5;
	public static final int STOP_DELAY = 3000;
	public static final int OVERALL_MOTION_TIME = 1500;
	public static final int UPWARD = 1;
	public static final int UNKNOWN = 0;
	public static final int DOWNWARD = -1;
	
	public static final String FONT_NAME = "ARIAL";

	private static final Color DOORS_COLOR = new Color( 192, 192, 255 );
	private static final Color LIFT_COLOR = new Color( 128, 128, 255 );
	private static final Color TEXT_COLOR = Color.WHITE;

	private EventManager eventManager;

	private MainForm mainForm;

	private TaskList upperList;
	private TaskList lowerList;
	
	private int motionDirection;
	
	private double curFloor;
	private double doorsState;
	
	boolean wasBreakage;

	/**
	 * Default constructor.
	 *
	 */
	public Lift()
	{
		upperList = new TaskList();
		lowerList = new TaskList();
		
		motionDirection = UPWARD;
		curFloor = 1.0;
		doorsState = 0.0;
		
		wasBreakage = false;
	}
	
	/**
	 * Chooses direction.
	 * @param context Recieved context
	 * @unimod.action.descr    .
	 */
	public void chooseDirection( StateMachineContext context )
	{
		//print info
		mainForm.printInfo( ":  ." );
		
		Event event;
		
		if ( upperList.isEmpty() && lowerList.isEmpty() )
			return;
		
		if ( motionDirection == UPWARD )
			event = new Event( LiftEventProvider.seGoUp );
		else
			event = new Event( LiftEventProvider.seGoDown );
		eventManager.handle( event, StateMachineContextImpl.create() );
	}
	
	/**
	 * Changes direction of lift movement (DOWNWARD to UPWARD and vice versa).
	 * @param context Recieved context
	 * @unimod.action.descr    .
	 */
	public void changeDirection( StateMachineContext context )
	{
		if ( motionDirection == UPWARD )
			motionDirection = DOWNWARD;
		else
			motionDirection = UPWARD;
	}
	
	/**
	 * Makes lift moving up.
	 * @param context Recieved context
	 * @unimod.action.descr   .
	 */
	public void goUp( StateMachineContext context )
	{
		int minFloor = Integer.MAX_VALUE;
		boolean ok = false;
		for ( int i = 0; i < upperList.size(); i++ )
			if ( ( upperList.itemFloor(i) >= curFloor ) && ( upperList.itemFloor(i) < minFloor ) )
			{
				ok = true;
				minFloor = upperList.itemFloor(i); 
			}
		if ( ( !ok ) && ( upperList.size() > 0 ) )
		{
			ok = true;
			minFloor = upperList.itemFloor(0);
		}
			
		if ( ok )
		{
			//print info
			if ( minFloor > curFloor )
				mainForm.printInfo( ":  ." );
			else if ( minFloor < curFloor )
				mainForm.printInfo( ":  ." );

			double destFloor = getFloorConsBreak( curFloor, minFloor );
			moveToFloor( destFloor );
			if ( destFloor != minFloor )
			{
				Event e = new Event( LiftEventProvider.seHasBroken );
				eventManager.handle( e, StateMachineContextImpl.create() );
			}
			else
			{
				upperList.removeTask( minFloor );
				try
				{
					Thread.sleep( 500 );
				}
				catch( Exception e )
				{
				}
				
				Event e = new Event( LiftEventProvider.seOpenTheDoors );
				eventManager.handle( e, StateMachineContextImpl.create() );
			}
		}
		else
		{
			Event event = new Event( LiftEventProvider.seQueueIsEmpty );
			eventManager.handle( event, StateMachineContextImpl.create() );
		}
	}
	
	/**
	 * Makes lift moving down.
	 * @param context Recieved context
	 * @unimod.action.descr   .
	 */
	public void goDown( StateMachineContext context )
	{
		int maxFloor = Integer.MIN_VALUE;
		boolean ok = false;
		for ( int i = 0; i < lowerList.size(); i++ )
			if ( ( lowerList.itemFloor(i) <= curFloor ) && ( lowerList.itemFloor(i) > maxFloor ) )
			{
				ok = true;
				maxFloor = lowerList.itemFloor(i); 
			}
		if ( ( !ok ) && ( lowerList.size() > 0 ) )
		{
			ok = true;
			maxFloor = lowerList.itemFloor(0);
		}
			
		if ( ok )
		{
			//print info
			if ( maxFloor > curFloor )
				mainForm.printInfo( ":  ." );
			else if ( maxFloor < curFloor )
				mainForm.printInfo( ":  ." );

			double destFloor = getFloorConsBreak( curFloor, maxFloor );
			moveToFloor( destFloor );
			if ( destFloor != maxFloor )
			{
				Event e = new Event( LiftEventProvider.seHasBroken );
				eventManager.handle( e, StateMachineContextImpl.create() );
			}
			else
			{
				lowerList.removeTask( maxFloor );
				try
				{
					Thread.sleep( 500 );
				}
				catch( Exception e )
				{
				}
				
				Event e = new Event( LiftEventProvider.seOpenTheDoors );
				eventManager.handle( e, StateMachineContextImpl.create() );
			}
		}
		else
		{
			Event event = new Event( LiftEventProvider.seQueueIsEmpty );
			eventManager.handle( event, StateMachineContextImpl.create() );
		}
	}

	
	/**
	 * Makes lift opening the doors.
	 * @param context Recieved context
	 * @unimod.action.descr   .
	 */
	public void openTheDoors( StateMachineContext context )
	{
		//print info
		mainForm.printInfo( ":  ." );
		
		for ( int i = 0; i <= 50; i++ )
		{
			doorsState = i * 0.02;
			drawLift();
			try
			{
				Thread.sleep( 5 );
			}
			catch( Exception e )
			{
			}
		}
	}
	
	/**
	 * Makes lift waiting for some time before doors closing.
	 * @param context Recieved context
	 * @unimod.action.descr     .
	 */ 
	public void waitOpened( StateMachineContext context )
	{
		//print info
		mainForm.printInfo( ":    ." );

		try
		{
			Thread.sleep( STOP_DELAY );
		}
		catch( Exception e )
		{
		}
		
    	Event event = new Event( LiftEventProvider.seCloseTheDoors );
		eventManager.handle( event, StateMachineContextImpl.create() );
	}
	
	/**
	 * Makes lift closing the doors.
	 * @param context Recieved context
	 * @unimod.action.descr   .
	 */
	public void closeTheDoors( StateMachineContext context )
	{
		//print info
		mainForm.printInfo( ":  ." );
		
		for ( int i = 50; i >= 0; i-- )
		{
			doorsState = i * 0.02;
			drawLift();
			try
			{
				Thread.sleep( 5 );
			}
			catch( Exception e )
			{
			}
		}
	}
	
	/**
	 * Obtains event manager.
	 * @param manager Event manager
	 */
	public void setEventManager( EventManager manager )
	{
		eventManager = manager;
	}
	
	/**
	 * Obtains main form
	 * @param f Form
	 */
	public void setForm( MainForm f )
	{
		mainForm = f;
	}
	
	/**
	 * Adds a new task to the queue.
	 * @param floor Number of a floor
	 * @param direction Direction of the request (UPWARD, DOWNWARD or UNKNOWN)
	 * @param outCall Flag that represents outer call
	 */
	public void addNewTask( int floor, int direction, boolean outCall )
	{
		boolean added = false;
		if ( ( direction == UPWARD ) && ( !upperList.contains( floor, outCall ) ) )
		{
			upperList.add( floor, outCall );
			added = true;
			
			//print info
			mainForm.printInfo( "    " + String.valueOf( floor) + " .    ." );
		}
		else if ( ( direction == DOWNWARD ) && ( !lowerList.contains( floor, outCall ) ) )
		{
			lowerList.add( floor, outCall );
			added = true;

			//print info
			mainForm.printInfo( "    " + String.valueOf( floor) + " .    ." );
		}
		else if ( ( direction == UNKNOWN ) && ( !lowerList.contains( floor, outCall ) ) )
		{
			if ( floor >= curFloor )
				upperList.add( floor, outCall );
			else
				lowerList.add( floor, outCall );
			added = true;
			
			//print info
			mainForm.printInfo( "   .     " + String.valueOf( floor ) + " ." );
		}
		
		if ( added )
		{
			Event event = new Event( LiftEventProvider.seNewTaskAdded );
			eventManager.handle( event, StateMachineContextImpl.create() );
		}
	}
	
	/**
	 * Draws lift.
	 *
	 */
	public void drawLift()
	{
		Rectangle rect = mainForm.getLiftRectangle();
		BufferedImage bufImage = new BufferedImage( rect.width, rect.height, BufferedImage.TYPE_3BYTE_BGR );
		Graphics g = bufImage.getGraphics();
		
		drawWell( g, rect );
		
		int x0 = rect.x + rect.width / 2;

		int y0 = rect.y + rect.height - 15;
		int dy = rect.height - 20;
		
		int lHeight = dy / FLOORS_COUNT;
		int lWidth = lHeight * 2 / 3;
		
		String floorText = ": " + String.valueOf( Math.floor( curFloor * 10 ) / 10 );

		Rectangle liftBody = new Rectangle( x0 - lWidth / 2, (int)( y0 - lHeight * curFloor ), lWidth, lHeight );
		Rectangle leftDoor = new Rectangle( (int)( x0 - lWidth * 0.5 * ( 1 + doorsState ) + 1), (int)( y0 - lHeight * ( curFloor - 0.1 ) ), lWidth / 2, (int)( lHeight * 0.8 ) );
		Rectangle rightDoor = new Rectangle( (int)( x0 + lWidth * 0.5 * doorsState + 1), (int)( y0 - lHeight * ( curFloor - 0.1 ) ), lWidth / 2, (int)( lHeight * 0.8 ) );
		Rectangle textRect = new Rectangle( rect.x, (int)( rect.y + rect.height - 10 ), floorText.length() * 8, 10 );
		
		//draw lift body 
		g.setColor( LIFT_COLOR );
		g.fillRect( liftBody.x, liftBody.y, liftBody.width, liftBody.height );
		g.setColor( Color.black );
		g.drawRect( liftBody.x, liftBody.y, liftBody.width, liftBody.height );

		//draw left door
		g.setColor( DOORS_COLOR );
		g.fillRect( leftDoor.x, leftDoor.y, leftDoor.width, leftDoor.height );
		g.setColor( Color.black );
		g.drawRect( leftDoor.x, leftDoor.y, leftDoor.width, leftDoor.height );

		//draw right door
		g.setColor( DOORS_COLOR );
		g.fillRect( rightDoor.x, rightDoor.y, rightDoor.width, rightDoor.height );
		g.setColor( Color.black );
		g.drawRect( rightDoor.x, rightDoor.y, rightDoor.width, rightDoor.height );
		
		g.setColor( TEXT_COLOR );
		g.setFont( new Font( FONT_NAME, Font.BOLD, 16 ) );
		g.drawString( floorText, textRect.x, textRect.y );
		
		Graphics liftGraphics = mainForm.getLiftGraphics();
		liftGraphics.drawImage( bufImage, 0, 0, rect.width, rect.height, mainForm );
	}
	
	/**
	 * Shows specified button highlighting
	 * @param floor Number of the floor
	 * @param direction Direction of the request
	 * @param outCall Flag that represents outer call
	 * @return Returns true if specified button is highlighted and false otherwise 
	 */
	public boolean isButtonHighlighted( int floor, int direction, boolean outCall )
	{
		boolean up = upperList.contains( floor, outCall );
		boolean down = lowerList.contains( floor, outCall ); 
		if ( direction == UPWARD )
			return up;
		else if ( direction == DOWNWARD )
			return down;
		else 
			return ( up | down );
	}
	
	private void drawWell( Graphics g, Rectangle rect )
	{
		int y = rect.y + rect.height - 15;
		int dy = ( rect.height - 20 ) / FLOORS_COUNT;
		int dx = dy * 2 / 3;

		g.setColor( TEXT_COLOR );
		
		for ( int i = 1; i < FLOORS_COUNT; i++ )
		{
			g.drawLine( rect.x, y - dy * i, rect.x + rect.width, y - dy * i );
		}
		
		g.drawLine( rect.x + rect.width / 2 - dx / 2, y - dy * FLOORS_COUNT, rect.x + rect.width / 2 - dx / 2, y );
		g.drawLine( rect.x + rect.width / 2 + dx / 2, y - dy * FLOORS_COUNT, rect.x + rect.width / 2 + dx / 2, y );
	}
	
	private void moveToFloor( double f )
	{
		double delta = f - curFloor;
		double floor = curFloor;
		
		int steps = 200;
		int time = OVERALL_MOTION_TIME / steps;
		double iter = Math.max( 10, (int)Math.abs( delta * steps / Lift.FLOORS_COUNT ) );
		
		for ( int i = 0; i <= iter; i++ )
		{
			curFloor = floor + delta * i / iter;
			drawLift();
			try
			{
				Thread.sleep( time );
			}
			catch( Exception e )
			{				
			}
		}
	}
	
	private double getFloorConsBreak( double fromFloor, double toFloor )
	{
		if ( wasBreakage )
		{	
			wasBreakage = false;
			return toFloor;
		}
		
		Random rand = new Random( System.currentTimeMillis() );
		if ( rand.nextBoolean() && rand.nextBoolean() )
		{
			wasBreakage = true;
			return fromFloor + ( toFloor - fromFloor ) * ( Math.abs( rand.nextInt( 9 ) ) + 1) / 10;
		}
			
		return toFloor;
	}
}
